from abc import abstractmethod
from datetime import datetime
from panther.logger import logger
from panther.configs import config
from panther.db.models import BaseUser
from panther.request import Request
from panther.cli.utils import import_error
from panther.exceptions import AuthenticationException
try:
    from jose import JWTError, jwt
except ImportError:
    import_error('python-jose')
    exit()


class BaseAuthentication:
    @classmethod
    @abstractmethod
    def authentication(cls, request: Request):
        """
        Return User Instance
        """
        raise cls.exception(f'{cls.__name__}.authentication() is not implemented.')

    @staticmethod
    def exception(message: str, /):
        logger.error(f'Authentication Error: "{message}"')
        return AuthenticationException


class JWTAuthentication(BaseAuthentication):
    model = BaseUser
    keyword = 'Bearer'
    algorithm = 'HS256'
    HTTP_HEADER_ENCODING = 'iso-8859-1'  # Header encoding (see RFC5987)

    @staticmethod
    def get_authorization_header(request: Request):
        auth = request.headers.authorization
        if isinstance(auth, str):
            auth = auth.encode(JWTAuthentication.HTTP_HEADER_ENCODING)
        return auth

    @classmethod
    def authentication(cls, request: Request):
        auth = cls.get_authorization_header(request).split()
        if not auth or auth[0].lower() != cls.keyword.lower().encode():
            raise cls.exception('token keyword is not valid')
        if len(auth) != 2:
            raise cls.exception('token should have 2 part')

        try:
            token = auth[1].decode()
        except UnicodeError:
            raise cls.exception('Unicode Error')

        payload = cls.decode_jwt(token)
        return cls.get_user(payload)

    @classmethod
    def get_user(cls, payload: dict):
        """
        Get UserModel from config else use default UserModel from cls.model
        """
        if (user_id := payload.get('user_id')) is None:
            raise cls.exception('Payload does not have user_id')

        user_model = config['user_model'] or cls.model
        if user := user_model.find_one(id=user_id):
            return user

        raise cls.exception('User not found')

    @classmethod
    def encode_jwt(cls, user_id: int) -> str:
        """Encode JWT from user_id."""
        expire = datetime.utcnow() + config['jwt_config'].life_time
        access_payload = {
            'token_type': 'access',
            'user_id': user_id,
            'exp': expire,
        }
        return jwt.encode(access_payload, config['jwt_config'].key, algorithm=config['jwt_config'].algorithm)

    @classmethod
    def decode_jwt(cls, token: str) -> dict:
        """Decode JWT token to user_id (it can return multiple variable ... )"""
        try:
            return jwt.decode(
                token=token,
                key=config['jwt_config'].key,
                algorithms=[config['jwt_config'].algorithm]
            )
        except JWTError as e:
            raise cls.exception(e)

    @classmethod
    def login(cls, user_id: int) -> str:
        """alias of encode_jwt()"""
        return cls.encode_jwt(user_id=user_id)

    @staticmethod
    def logout(user_id: int):
        # TODO:
        #   1. Save the token in cache
        #   2. Check logout cached token in authentication()
        pass

    @staticmethod
    def exception(message: str | JWTError, /):
        logger.error(f'JWT Authentication Error: "{message}"')
        return AuthenticationException




import os
from collections import deque

from rich.console import Group
from rich.layout import Layout
from rich.align import Align
from rich.panel import Panel
from rich.table import Table
from watchfiles import watch
from rich.live import Live
from rich import box


def monitor() -> None:
    monitoring_log_file = 'logs/monitoring.log'

    def _generate_table(rows: deque):
        layout = Layout()

        rows = list(rows)
        _, lines = os.get_terminal_size()

        table = Table(box=box.MINIMAL_DOUBLE_HEAD)
        table.add_column('Datetime', justify='center', style='magenta', no_wrap=True)
        table.add_column('Method', justify='center', style='cyan')
        table.add_column('Path', justify='center', style='cyan')
        table.add_column('Client', justify='center', style='cyan')
        table.add_column('Response Time', justify='center', style='blue')
        table.add_column('Status Code', justify='center', style='blue')

        for row in rows[-lines:]:  # It will give us "lines" last lines of "rows"
            table.add_row(*row)
        layout.update(table)

        return Panel(
            Align.center(Group(table)),
            box=box.ROUNDED,
            padding=(1, 2),
            title='Monitoring',
            border_style='bright_blue',
        )

    with open(monitoring_log_file) as f:
        f.readlines()  # Set cursor at the end of file

        _, init_lines_count = os.get_terminal_size()
        messages = deque(maxlen=init_lines_count - 10)  # Save space for header and footer

        with Live(_generate_table(messages), auto_refresh=False, vertical_overflow='visible', screen=True) as live:
            try:
                for _ in watch(monitoring_log_file):
                    data = f.readline().split('|')
                    # 2023-03-24 01:42:52 | GET | /user/317/ | 127.0.0.1:48856 |  0.0366 ms | 200
                    messages.append(data)
                    live.update(_generate_table(messages))
                    live.refresh()
            except KeyboardInterrupt:
                pass



from panther.app import API
from panther.configs import config


@API()
async def list_models():
    for i, m in enumerate(config['models']):
        m.pop('class')
        m['app'] = '.'.join(a for a in m['app'])
        m['index'] = i
    return config['models']


@API()
async def model_list(index: int):
    model = config['models'][index]['class']
    return model.find()


@API()
async def model_retrieve(index: int, id: int):
    model = config['models'][index]['class']
    return model.find_one(id=id)


